# Notifier
#
# Copyright (C) 2022 Pataar <me@pataar.nl>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations

import apprise
import logging
import pathlib
import re
from ..common import JobEvent, RequestType

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Dict,
    Any,
    List,
)

if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    from ..common import WebRequest
    from .file_manager.file_manager import FileManager
    from .klippy_apis import KlippyAPI as APIComp

class Notifier:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.notifiers: Dict[str, NotifierInstance] = {}
        self.events: Dict[str, List[NotifierInstance]] = {}
        prefix_sections = config.get_prefix_sections("notifier")
        self.register_remote_actions()
        for section in prefix_sections:
            cfg = config[section]
            try:
                notifier = NotifierInstance(cfg)
                for job_event in list(JobEvent):
                    if job_event == JobEvent.STANDBY:
                        continue
                    evt_name = str(job_event)
                    if "*" in notifier.events or evt_name in notifier.events:
                        self.events.setdefault(evt_name, []).append(notifier)
                logging.info(f"Registered notifier: '{notifier.get_name()}'")
            except Exception as e:
                msg = f"Failed to load notifier[{cfg.get_name()}]\n{e}"
                self.server.add_warning(msg, exc_info=e)
                continue
            self.notifiers[notifier.get_name()] = notifier

        self.register_endpoints(config)
        self.server.register_event_handler(
            "job_state:state_changed", self._on_job_state_changed
        )

    def register_remote_actions(self):
        self.server.register_remote_method("notify", self.notify_action)

    async def notify_action(self, name: str, message: str = ""):
        if name not in self.notifiers:
            raise self.server.error(f"Notifier '{name}' not found", 404)
        notifier = self.notifiers[name]
        await notifier.notify("remote_action", [], message)

    async def _on_job_state_changed(
            self,
            job_event: JobEvent,
            prev_stats: Dict[str, Any],
            new_stats: Dict[str, Any]
    ) -> None:
        evt_name = str(job_event)
        for notifier in self.events.get(evt_name, []):
            await notifier.notify(evt_name, [prev_stats, new_stats])

    def register_endpoints(self, config: ConfigHelper):
        self.server.register_endpoint(
            "/server/notifiers/list", RequestType.GET, self._handle_notifier_list
        )
        self.server.register_debug_endpoint(
            "/debug/notifiers/test", RequestType.POST, self._handle_notifier_test
        )

    async def _handle_notifier_list(
        self, web_request: WebRequest
    ) -> Dict[str, Any]:
        return {"notifiers": self._list_notifiers()}

    def _list_notifiers(self) -> List[Dict[str, Any]]:
        return [notifier.as_dict() for notifier in self.notifiers.values()]

    async def _handle_notifier_test(
        self, web_request: WebRequest
    ) -> Dict[str, Any]:

        name = web_request.get_str("name")
        if name not in self.notifiers:
            raise self.server.error(f"Notifier '{name}' not found", 404)
        notifier = self.notifiers[name]

        kapis: APIComp = self.server.lookup_component('klippy_apis')
        result: Dict[str, Any] = await kapis.query_objects(
            {'print_stats': None}, default={})
        print_stats = result.get('print_stats', {})
        print_stats["filename"] = "notifier_test.gcode"  # Mock the filename

        await notifier.notify(notifier.events[0], [print_stats, print_stats])

        return {
            "status": "success",
            "stats": print_stats
        }

class NotifierInstance:
    def __init__(self, config: ConfigHelper) -> None:
        self.config = config
        name_parts = config.get_name().split(maxsplit=1)
        if len(name_parts) != 2:
            raise config.error(f"Invalid Section Name: {config.get_name()}")
        self.server = config.get_server()
        self.name = name_parts[1]
        self.apprise = apprise.Apprise()
        self.attach = config.gettemplate("attach", None)
        url_template = config.gettemplate("url")
        self.url = url_template.render()

        if re.match(r"\w+?://", self.url) is None:
            raise config.error(f"Invalid url for: {config.get_name()}")

        self.title = config.gettemplate("title", None)
        self.body = config.gettemplate("body", None)
        upper_body_format = config.get("body_format", 'text').upper()
        if not hasattr(apprise.NotifyFormat, upper_body_format):
            raise config.error(f"Invalid body_format for {config.get_name()}")
        self.body_format = getattr(apprise.NotifyFormat, upper_body_format)
        self.events: List[str] = config.getlist("events", separator=",")
        self.apprise.add(self.url)

    def as_dict(self):
        return {
            "name": self.name,
            "url": self.config.get("url"),
            "title": self.config.get("title", None),
            "body": self.config.get("body", None),
            "body_format": self.config.get("body_format", None),
            "events": self.events,
            "attach": self.attach
        }

    async def notify(
        self, event_name: str, event_args: List, message: str = ""
    ) -> None:
        context = {
            "event_name": event_name,
            "event_args": event_args,
            "event_message": message
        }

        rendered_title = (
            '' if self.title is None else self.title.render(context)
        )
        rendered_body = (
            event_name if self.body is None else self.body.render(context)
        )

        # Verify the attachment
        attachments: List[str] = []
        if self.attach is not None:
            fm: FileManager = self.server.lookup_component("file_manager")
            try:
                rendered = self.attach.render(context)
            except self.server.error as e:
                self.server.add_warning(
                    f"[notifier {self.name}]: The attachment is not valid. The "
                    "template failed to render.",
                    f"notifier {self.name}",
                    exc_info=e
                )
                self.attach = None
            else:
                for item in rendered.splitlines():
                    item = item.strip()
                    if not item:
                        continue
                    if re.match(r"https?://", item) is not None:
                        # Attachment is a url, system check not necessary
                        attachments.append(item)
                        continue
                    attach_path = pathlib.Path(item).expanduser().resolve()
                    if not attach_path.is_file():
                        self.server.add_warning(
                            f"[notifier {self.name}]: Invalid attachment detected, "
                            f"file does not exist: {attach_path}.",
                            f"notifier {self.name}"
                        )
                    elif not fm.can_access_path(attach_path):
                        self.server.add_warning(
                            f"[notifier {self.name}]: Invalid attachment detected, "
                            f"no read permission for the file {attach_path}.",
                            f"notifier {self.name}"
                        )
                    else:
                        attachments.append(str(attach_path))
        await self.apprise.async_notify(
            rendered_body.strip(), rendered_title.strip(),
            body_format=self.body_format,
            attach=None if not attachments else attachments
        )

    def get_name(self) -> str:
        return self.name


def load_component(config: ConfigHelper) -> Notifier:
    return Notifier(config)



# General Server Utilities
#
# Copyright (C) 2023 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import importlib.resources as ilr
import pathlib
import sys
import site
import re
import json
import logging
from dataclasses import dataclass
from importlib_metadata import Distribution, PathDistribution, PackageMetadata
from .exceptions import ServerError

# Annotation imports
from typing import (
    Optional,
    Dict,
    Any
)

def package_path() -> pathlib.Path:
    return pathlib.Path(__file__).parent.parent

def source_path() -> pathlib.Path:
    return package_path().parent

def is_git_repo(src_path: Optional[pathlib.Path] = None) -> bool:
    if src_path is None:
        src_path = source_path()
    return src_path.joinpath(".git").is_dir()

def find_git_repo(src_path: Optional[pathlib.Path] = None) -> Optional[pathlib.Path]:
    if src_path is None:
        src_path = source_path()
    if src_path.joinpath(".git").is_dir():
        return src_path
    for parent in src_path.parents:
        if parent.joinpath(".git").is_dir():
            return parent
    return None

def is_dist_package(item_path: Optional[pathlib.Path] = None) -> bool:
    """
    Check if the supplied path exists within a python dist installation or
    site installation.
    """
    if item_path is None:
        # Check Moonraker's package path
        item_path = package_path()
        if hasattr(site, "getsitepackages"):
            # The site module is present, get site packages for Moonraker's venv.
            # This is more "correct" than the fallback method.
            for site_dir in site.getsitepackages():
                site_path = pathlib.Path(site_dir)
                try:
                    if site_path.samefile(item_path.parent):
                        return True
                except Exception:
                    pass
    # Make an assumption based on the item and/or its parents.  If a folder
    # is named site-packages or dist-packages then presumably it is an
    # installed package
    if item_path.name in ("dist-packages", "site-packages"):
        return True
    for parent in item_path.parents:
        if parent.name in ("dist-packages", "site-packages"):
            return True
    return False

def package_version() -> Optional[str]:
    try:
        import moonraker.__version__ as ver  # type: ignore
        version = ver.__version__
    except Exception:
        pass
    else:
        if version:
            return version
    return None

def read_asset(asset_name: str) -> Optional[str]:
    if sys.version_info < (3, 10):
        with ilr.path("moonraker.assets", asset_name) as p:
            if not p.is_file():
                return None
            return p.read_text()
    else:
        files = ilr.files("moonraker.assets")
        with ilr.as_file(files.joinpath(asset_name)) as p:
            if not p.is_file():
                return None
            return p.read_text()

def get_asset_path() -> Optional[pathlib.Path]:
    if sys.version_info < (3, 10):
        with ilr.path("moonraker.assets", "__init__.py") as p:
            asset_path = p.parent
    else:
        files = ilr.files("moonraker.assets")
        with ilr.as_file(files.joinpath("__init__.py")) as p:
            asset_path = p.parent
    if not asset_path.is_dir():
        # Somehow running in a zipapp.  This is an error.
        return None
    return asset_path

def _load_release_info_json(dist_info: Distribution) -> Optional[Dict[str, Any]]:
    files = dist_info.files
    if files is None:
        return None
    for dist_file in files:
        if dist_file.parts[0] in ["..", "/"]:
            continue
        if dist_file.name == "release_info":
            pkg = dist_file.parts[0]
            logging.info(f"Package {pkg}: Detected release_info json file")
            try:
                return json.loads(dist_file.read_text())
            except Exception:
                logging.exception(f"Failed to load release_info from {dist_file}")
    return None

def _load_direct_url_json(dist_info: Distribution) -> Optional[Dict[str, Any]]:
    ret: Optional[str] = dist_info.read_text("direct_url.json")
    if ret is None:
        return None
    try:
        direct_url: Dict[str, Any] = json.loads(ret)
    except json.JSONDecodeError:
        return None
    return direct_url

def normalize_project_name(name: str) -> str:
    return re.sub(r"[-_.]+", "-", name).lower().replace('-', '_')

def load_distribution_info(
    venv_path: pathlib.Path, project_name: str
) -> PackageInfo:
    proj_name_normalized = normalize_project_name(project_name)
    site_items = venv_path.joinpath("lib").glob("python*/site-packages/")
    lib_paths = [str(p) for p in site_items if p.is_dir()]
    for dist_info in Distribution.discover(name=project_name, path=lib_paths):
        metadata = dist_info.metadata
        if metadata is None:
            continue
        if not isinstance(dist_info, PathDistribution):
            logging.info(f"Project {dist_info.name} not a PathDistribution")
            continue
        metaname = normalize_project_name(metadata["Name"] or "")
        if metaname != proj_name_normalized:
            continue
        release_info = _load_release_info_json(dist_info)
        install_info = _load_direct_url_json(dist_info)
        return PackageInfo(
            dist_info, metadata, release_info, install_info
        )
    raise ServerError(f"Failed to find distribution info for project {project_name}")

def is_vitualenv_project(
    venv_path: Optional[pathlib.Path] = None,
    pkg_path: Optional[pathlib.Path] = None,
    project_name: str = "moonraker"
) -> bool:
    if venv_path is None:
        venv_path = pathlib.Path(sys.exec_prefix)
    if pkg_path is None:
        pkg_path = package_path()
    if not pkg_path.exists():
        return False
    try:
        pkg_info = load_distribution_info(venv_path, project_name)
    except Exception:
        return False
    site_path = pathlib.Path(str(pkg_info.dist_info.locate_file("")))
    for parent in pkg_path.parents:
        try:
            if site_path.samefile(parent):
                return True
        except Exception:
            pass
    return True

@dataclass(frozen=True)
class PackageInfo:
    dist_info: Distribution
    metadata: PackageMetadata
    release_info: Optional[Dict[str, Any]]
    direct_url_data: Optional[Dict[str, Any]]
